@@ -1,3 +1,4 @@
+import bisect
 from abc import ABC, abstractmethod
 from copy import deepcopy
 from typing import Any, List
@@ -5,11 +6,13 @@ from typing import Any, List
 import numpy as np
 import pandas as pd
 
-from .transformations.datacube_cyclic.cyclic_axis_decorator import cyclic
-from .transformations.datacube_mappers.mapper_axis_decorator import mapper
-from .transformations.datacube_merger.merger_axis_decorator import merge
-from .transformations.datacube_reverse.reverse_axis_decorator import reverse
-from .transformations.datacube_type_change.type_change_axis_decorator import type_change
+from .transformations.datacube_cyclic.datacube_cyclic import DatacubeAxisCyclic
+from .transformations.datacube_mappers.datacube_mappers import DatacubeMapper
+from .transformations.datacube_merger.datacube_merger import DatacubeAxisMerger
+from .transformations.datacube_reverse.datacube_reverse import DatacubeAxisReverse
+from .transformations.datacube_type_change.datacube_type_change import (
+    DatacubeAxisTypeChange,
+)
 
 
 class DatacubeAxis(ABC):
@@ -19,10 +22,12 @@ class DatacubeAxis(ABC):
     reorder = False
     type_change = False
 
-    def update_axis(self):
-        if self.is_cyclic:
-            self = cyclic(self)
-        return self
+    def order_tranformations(self):
+        self.transformations = sorted(self.transformations, key=lambda x: transformations_order[type(x)])
+
+    def give_transformations_parents(self):
+        for i, transform in enumerate(self.transformations[1:]):
+            transform.parent = self.transformations[i - 1]
 
     # Convert from user-provided value to CONTINUOUS type (e.g. float, pd.timestamp)
     @abstractmethod
@@ -43,15 +48,21 @@ class DatacubeAxis(ABC):
         pass
 
     def to_intervals(self, range):
-        return [range]
+        intervals = [range]
+        for transformation in self.transformations[::-1]:
+            intervals = transformation.to_intervals(range, intervals, self)
+        return intervals
 
     def remap(self, range: List) -> Any:
-        return [range]
+        ranges = [range]
+        for transformation in self.transformations[::-1]:
+            ranges = transformation.remap(range, ranges, self)
+        return ranges
 
     def unmap_to_datacube(self, path, unmapped_path):
         return (path, unmapped_path)
 
-    def find_indexes(self, path, datacube):
+    def find_standard_indexes(self, path, datacube):
         unmapped_path = {}
         path_copy = deepcopy(path)
         for key in path_copy:
@@ -60,46 +71,75 @@ class DatacubeAxis(ABC):
         subarray = datacube.select(path, unmapped_path)
         return datacube.datacube_natural_indexes(self, subarray)
 
+    def find_indexes(self, path, datacube):
+        indexes = self.find_standard_indexes(path, datacube)
+        for transformation in self.transformations[::-1]:
+            indexes = transformation.find_modified_indexes(indexes, path, datacube, self)
+        return indexes
+
     def offset(self, value):
-        return 0
+        offset = 0
+        for transformation in self.transformations[::-1]:
+            offset = transformation.offset(value, self, offset)
+        return offset
 
     def unmap_path_key(self, key_value_path, leaf_path, unwanted_path):
+        for transformation in self.transformations[::-1]:
+            (key_value_path, leaf_path, unwanted_path) = transformation.unmap_path_key(
+                key_value_path, leaf_path, unwanted_path, self
+            )
         return (key_value_path, leaf_path, unwanted_path)
 
+    def unmap_tree_node(self, node, unwanted_path):
+        for transformation in self.transformations[::-1]:
+            (node, unwanted_path) = transformation.unmap_tree_node(node, unwanted_path)
+        return (node, unwanted_path)
+
     def _remap_val_to_axis_range(self, value):
+        for transformation in self.transformations[::-1]:
+            value = transformation._remap_val_to_axis_range(value, self)
         return value
 
-    def find_indices_between(self, index_ranges, low, up, datacube, method=None):
-        # TODO: add method for snappping
+    def find_standard_indices_between(self, indexes, low, up, datacube, method=None):
         indexes_between_ranges = []
-        for indexes in index_ranges:
-            if self.name in datacube.complete_axes:
-                # Find the range of indexes between lower and upper
-                # https://pandas.pydata.org/docs/reference/api/pandas.Index.searchsorted.html
-                # Assumes the indexes are already sorted (could sort to be sure) and monotonically increasing
-                if method == "surrounding" or method == "nearest":
-                    start = indexes.searchsorted(low, "left")
-                    end = indexes.searchsorted(up, "right")
-                    start = max(start - 1, 0)
-                    end = min(end + 1, len(indexes))
-                    indexes_between = indexes[start:end].to_list()
-                    indexes_between_ranges.append(indexes_between)
-                else:
-                    start = indexes.searchsorted(low, "left")
-                    end = indexes.searchsorted(up, "right")
-                    indexes_between = indexes[start:end].to_list()
-                    indexes_between_ranges.append(indexes_between)
+
+        if self.name in datacube.complete_axes and self.name not in datacube.transformed_axes:
+            # Find the range of indexes between lower and upper
+            # https://pandas.pydata.org/docs/reference/api/pandas.Index.searchsorted.html
+            # Assumes the indexes are already sorted (could sort to be sure) and monotonically increasing
+            if method == "surrounding" or method == "nearest":
+                start = indexes.searchsorted(low, "left")
+                end = indexes.searchsorted(up, "right")
+                start = max(start - 1, 0)
+                end = min(end + 1, len(indexes))
+                indexes_between = indexes[start:end].to_list()
+                indexes_between_ranges.extend(indexes_between)
             else:
-                if method == "surrounding" or method == "nearest":
-                    start = indexes.index(low)
-                    end = indexes.index(up)
-                    start = max(start - 1, 0)
-                    end = min(end + 1, len(indexes))
-                    indexes_between = indexes[start:end]
-                    indexes_between_ranges.append(indexes_between)
-                else:
-                    indexes_between = [i for i in indexes if low <= i <= up]
-                    indexes_between_ranges.append(indexes_between)
+                start = indexes.searchsorted(low, "left")
+                end = indexes.searchsorted(up, "right")
+                indexes_between = indexes[start:end].to_list()
+                indexes_between_ranges.extend(indexes_between)
+        else:
+            if method == "surrounding" or method == "nearest":
+                start = bisect.bisect_left(indexes, low)
+                end = bisect.bisect_right(indexes, up)
+                start = max(start - 1, 0)
+                end = min(end + 1, len(indexes))
+                indexes_between = indexes[start:end]
+                indexes_between_ranges.extend(indexes_between)
+            else:
+                lower_idx = bisect.bisect_left(indexes, low)
+                upper_idx = bisect.bisect_right(indexes, up)
+                indexes_between = indexes[lower_idx:upper_idx]
+                indexes_between_ranges.extend(indexes_between)
+        return indexes_between_ranges
+
+    def find_indices_between(self, indexes_ranges, low, up, datacube, method=None):
+        indexes_between_ranges = self.find_standard_indices_between(indexes_ranges, low, up, datacube, method)
+        for transformation in self.transformations[::-1]:
+            indexes_between_ranges = transformation.find_indices_between(
+                indexes_ranges, low, up, datacube, method, indexes_between_ranges, self
+            )
         return indexes_between_ranges
 
     @staticmethod
@@ -120,17 +160,25 @@ class DatacubeAxis(ABC):
             raise ValueError(f"Could not create a mapper for index type {values.dtype.type} for axis {name}")
 
 
-@reverse
-@cyclic
-@mapper
-@type_change
+transformations_order = [
+    DatacubeAxisMerger,
+    DatacubeAxisReverse,
+    DatacubeAxisCyclic,
+    DatacubeMapper,
+    DatacubeAxisTypeChange,
+]
+transformations_order = {key: i for i, key in enumerate(transformations_order)}
+
+
 class IntDatacubeAxis(DatacubeAxis):
     def __init__(self):
         self.name = None
         self.tol = 1e-12
         self.range = None
+        # TODO: Maybe here, store transformations as a dico instead
         self.transformations = []
         self.type = 0
+        self.can_round = True
 
     def parse(self, value: Any) -> Any:
         return float(value)
@@ -145,10 +193,6 @@ class IntDatacubeAxis(DatacubeAxis):
         return value
 
 
-@reverse
-@cyclic
-@mapper
-@type_change
 class FloatDatacubeAxis(DatacubeAxis):
     def __init__(self):
         self.name = None
@@ -156,6 +200,7 @@ class FloatDatacubeAxis(DatacubeAxis):
         self.range = None
         self.transformations = []
         self.type = 0.0
+        self.can_round = True
 
     def parse(self, value: Any) -> Any:
         return float(value)
@@ -170,7 +215,6 @@ class FloatDatacubeAxis(DatacubeAxis):
         return value
 
 
-@merge
 class PandasTimestampDatacubeAxis(DatacubeAxis):
     def __init__(self):
         self.name = None
@@ -178,6 +222,7 @@ class PandasTimestampDatacubeAxis(DatacubeAxis):
         self.range = None
         self.transformations = []
         self.type = pd.Timestamp("2000-01-01T00:00:00")
+        self.can_round = False
 
     def parse(self, value: Any) -> Any:
         if isinstance(value, np.str_):
@@ -200,7 +245,6 @@ class PandasTimestampDatacubeAxis(DatacubeAxis):
         return None
 
 
-@merge
 class PandasTimedeltaDatacubeAxis(DatacubeAxis):
     def __init__(self):
         self.name = None
@@ -208,6 +252,7 @@ class PandasTimedeltaDatacubeAxis(DatacubeAxis):
         self.range = None
         self.transformations = []
         self.type = np.timedelta64(0, "s")
+        self.can_round = False
 
     def parse(self, value: Any) -> Any:
         if isinstance(value, np.str_):
@@ -230,13 +275,13 @@ class PandasTimedeltaDatacubeAxis(DatacubeAxis):
         return None
 
 
-@type_change
 class UnsliceableDatacubeAxis(DatacubeAxis):
     def __init__(self):
         self.name = None
         self.tol = float("NaN")
         self.range = None
         self.transformations = []
+        self.can_round = False
 
     def parse(self, value: Any) -> Any:
         return value
@@ -258,6 +303,8 @@ _type_to_axis_lookup = {
     np.datetime64: PandasTimestampDatacubeAxis(),
     np.timedelta64: PandasTimedeltaDatacubeAxis(),
     np.float64: FloatDatacubeAxis(),
+    np.float32: FloatDatacubeAxis(),
+    np.int32: IntDatacubeAxis(),
     np.str_: UnsliceableDatacubeAxis(),
     str: UnsliceableDatacubeAxis(),
     np.object_: UnsliceableDatacubeAxis(),
